{$A+,B-,D-,E+,F+,G+,I-,L-,N+,P+,Q-,R-,S-,T-,V+,X+,Y-}
(****************************************************************************
 * GX 2 unit for working with time dependent programs. Very useful to       *
 * get the same results on every machine (slower computer, less fps, but    *
 * all displayments work in the same time on every machine)                 *
 * Copyright information in gx2.txt                                         *
 * Current version completely based upon the (FPC-)unit CPU (Rel0Beta9)     *
 * written by Thomas Schatzl. BP-Port written by Stefan Goehler.            *
 * Differences to CPU:                                                      *
 * * type timer renamed to ttimer, ztimer renamed to tztimer                *
 * * delay parameter is now given in miliseconds                            *
 * + procedure WaitFor added - start timer with start, do what you wanna do *
 *   and call this proc - it'll wait for the remaining time                 *
 * * after delay and waitfor, a reset is performed                          *
 ****************************************************************************)
unit timebase;

interface

{ constants for the capabilities field of the CPU records' capabilites field }

type
  dword = longint;

const cHAVE_FPU    = $0001;  { floating point unit }
      cHAVE_CPUID  = $0002;  { cpuid instruction supported }
      cHAVE_RDTSC  = $0004;  { rdtsc instruction supported }
      cHAVE_MMX    = $0008;  { MMX extensions available }
      cHAVE_EMMX   = $0010;  { Cyrix  extended MMX available }
      cHAVE_3DNOW  = $0020;  { AMD 3Dnow extensions supported }
      cHAVE_KNI    = $0040;  { Katmai New Instructions supported }
      cHAVE_ISSE   = $0040;	 { Internet Streaming SIMD Extionsions supported }
      						 { equals KNI }
      cHAVE_SERNUM = $0080;  { PIII+ serial number available }

{ cpu family identifier constants }

const cFAMILY_I386     = 3; { 80386 processor class }
      cFAMILY_I486     = 4; { 80486 processor class }
      cFAMILY_PENTIUM  = 5; { Pentium (80586) class processor }
      cFAMILY_PENTIUM2 = 6; { Pentium2 (80686) class processor }
      cFAMILY_UNKNOWN  = 7; { Unkown processor class }

{ manufacturer identifier constants (sorted alphabetically) }

const cMFC_AMD     = 0;
      cMFC_CENTAUR = 1;
      cMFC_CYRIX   = 2;
      cMFC_INTEL   = 3;
      cMFC_NEXGEN  = 4;
      cMFC_UMC     = 5;
      cMFC_OTHER   = 6;

{ the timing frequency of the programming interval timer }

const PIT_RATE = $1234DD;

{ strings of different CPU brands reported by the CPUid instruction. For
  reference only.
}
                         
const manufacturer_idstrings : array[cMFC_AMD..cMFC_UMC] of string[12] =
                             ('AuthenticAMD',
                              'CentaurHauls',
                              'CyrixInstead',
                              'GenuineIntel',
                              'NexGenDriven',
                              'UMC UMC UMC ');

{ list of known official ratings of CPUs. This list MUST be sorted in
  descending CPU speed order that these ratings are correctly identified.
  To add other ratings simply insert it at a proper place and adjust the array
  size accordingly.

  You DON'T need to change the procedure which does the
  rating calculation, additional ratings are automatically detected.

  It's best to contact me to insert it in the official CPU-release...

  BTW: This list SHOULD NOT contain ratings of overclocked CPU's, this array
  is purely for informational purposes. Always use the clock_rate entry of the
  CPU record instead of the rating for correct speed values
}

const known_ratings : array [0..35] of DWord =
                    (800, 750, 700, 666, 600, 533, 500, 450, 433, 400, 380,
                     366, 350, 333, 300, 266, 240, 233, 225, 200, 180, 166,
                     150, 133, 120, 100,  90,  75,  66,  50,  40,  33,  30,
                      25,  16,   0);

type serial_number = array[0..2] of DWord;

{ this record holds the various CPU characteristics detected by this unit }

type cpu_rec = record
             cpu_id : string[12];    { the string returned by CPUid, else 'Unknown'}
             family : Byte;          { the CPU family }
             model : Byte;           { CPU model, 0xFF means unknown }
             stepping : Byte;        { CPU stepping, 0xFF means unknown }
             manufacturer : Byte;    { the CPU manufacturer }
             capabilities : DWord;   { capabilities bit field }
             clock_rate : DWord;     { detected clock rate in Mhz }
             rating : DWord;         { detected CPU rating in Mhz }
             sernum : serial_number; { processor specific serial number }
     end;

{ default cpu record, which is filled up at startup }

var this_cpu : cpu_rec;

{ this procedure fills out the CPU record which is given to it as a parameter }

procedure get_cpu_information(var c : cpu_rec);

{ returns the family string }

function get_family_string(f : DWord) : string;

{ returns the manufacturers name string }

function get_mfc_string(m : DWord) : string;

{ returns the processor serial number as a string }

function get_sernum_string(ser : serial_number) : string;

{ a helper function to obtain CPU information from the this_cpu variable more
  easily. If more than one capability wants to be checked, they have to be
  OR'ed.
}

function cpu_supported(capability : DWord) : Boolean;

{ returns the current value of the internal time stamp counter. There is no
  check wheter the rdtsc instruction is supported or not, so be sure to check
  it before calling this procedure.
}

procedure get_rdtsc_value(var val : comp);

{ returns the current value of the programmable interval timer. This result is
  only accurate if you DON'T fiddle around with the timer registers.
}

procedure get_pit_value(var val : comp);

{ base timer object which runs on any platform supported by FPC pascal and
  having the DOS unit. The accuracy of this timer is 1/100 of a second. This
  resolution is only valid on DOS platforms, on others this may vary.

  You are allowed to have as much timers timing at the same time as you want.
}

type pTimer = ^tTimer;
     tTimer = object
           constructor Init;
           destructor Done;

           { resets the timer }
           procedure Reset; virtual;
           { starts the timer }
           procedure Start; virtual;
           { stops the timer }
           procedure Stop; virtual;

           { returns the time passed between (subsequent) calls to the
             Start/Stop methods }
           function Count : Comp; virtual;
           { returns the time passed since the last Start call in addition to
             the time passed between earlier calls to the Start/Stop methods.
             If this procedure is called when the timer isn't timing the
             value returned may be undefined }
           function Lap : Comp; virtual;
           { returns the resolution of the timer in parts of a second }
           function Resolution : Real; virtual;

           { returns the current state of the timer }
           function IsTiming : Boolean; virtual;

           { high precision delay procedure. The delay time given as a
             parameter is in miliseconds. You MUST NOT let a timer itself
             time his own delay procedure, else unpredicable results may
             occur.The delayed time is automatically added to the elapsed
             time }
           procedure Delay(time : DWord); virtual;

           procedure WaitFor(time : DWord); virtual;
           private

           { reads the current value of the used system timers }
           function ReadTime : Comp; virtual;

           private

           { this flag reflects the current state of the timer }
           timing : Boolean;

           { last start time of the timer. These two values may be in another
             scale than the times returned by the various time result methods }
           tmStart,
           { elapsed time, for measuring subsequent Start/Stop sequences }
           tmElapsed : Comp;
     end;

{ Zen timer for timing periods of up to 280+ years with a microsecond accuracy
  using the internal time stamp counter on machines supporting it.
}

     pZTimer = ^tZTimer;
     tZTimer = object(tTimer)
            constructor Init;
 	    destructor Done;

            function Count : Comp; virtual;
            function Lap : Comp; virtual;
            function Resolution : Real; virtual;
            
            {protected}

            private
            function ReadTime : Comp; virtual;

            private
            { converts the internally used time values to microsecond scale }
            function TicksTo_usec(t : Comp) : Comp;
            private

            { this flag shows the method used by the ZenTimer }
            use_rdtsc : Boolean;
            { used to convert the internally used timing scale to microsecond
              values }
            rate : DWord;
     end;

{ Ultra long timer object. It's accuracy is about 1/18 seconds with a timing
  period of up to one day.
}

     pULZTimer = ^ULZTimer;
     ULZTimer = object(tTimer)
              constructor Init;
 	      destructor Done;

              function Resolution : Real; virtual;
            
{              protected}
            private

              function ReadTime : Comp; virtual;
     end;

implementation


{ used for the 1/100 second timer only }
uses dos;



function get_cpu_clockrate(family : Byte) : DWord; forward;
function get_cpu_rating(clock : DWord) : DWord; forward;

{ Manufacturer check
  A specific manufacturer is detected most of the time by comparing the cpuid
  strings with known ones (different are C6, here you can't rely on that the
  cpuid string is valid, because you can change it via some MSR's on your own)
}

{ CyrixInstead = 64616574 736E4978 69727943 }

function is_cyrix : Boolean; assembler;
asm
  db 66h;xor   ax,ax
  db 0fh,0A2h {cpuid}
  db 66h;cmp   ax,1;dw 0000h
         jl @LNoCyrix
  db 66h;sub   bx,7943h;dw 6972h
  db 66h;sub   dx,4978h;dw 736Eh
  db 66h;sub   cx,6574h;dw 6461h
  db 66h;or    bx,cx
  db 66h;or    bx,dx
  db 0Fh,94h,0C0h{setz al}
  jmp @LCyrixExit

  @LNoCyrix:
         mov  ax,0
  @LCyrixExit:
end;


{ AuthenticAMD = 41757468 656E7469 63414D44 }

function is_amd : Boolean; assembler;
asm
  db 66h;xor   ax,ax
  db 0fh,0A2h {cpuid}
  db 66h;cmp   ax,1;dw 0000h
         jl @LNoAmd
  db 66h;sub   bx,7541h;dw 6874h
  db 66h;sub   dx,6E65h;dw 6974h
  db 66h;sub   cx,4163h;dw 444Dh
  db 66h;or    bx,cx
  db 66h;or    bx,dx
  db 0Fh,94h,0C0h{setz al}
         jmp @LAmdExit
  @LNoAmd:
          mov  ax,0
  @LAmdExit:
end;

{ GenuineIntel = 47656E75 696E6549 6E74656C }

function is_intel : Boolean; assembler;
asm
  db 66h;xor   ax,ax
  db 0fh,0A2h {cpuid}
  db 66h;cmp   ax,0;dw 0000h
         je @LNoIntel
  db 66h;sub   bx,6547h;dw 756Eh
  db 66h;sub   dx,6E69h;dw 4965h
  db 66h;sub   cx,746Eh;dw 6C65h
  db 66h;or    bx,cx
  db 66h;or    bx,dx
  db 0Fh,94h,0C0h{setz al}
   jmp @LIntelExit
  @LNoIntel:
         mov  ax,0
  @LIntelExit:
end;

{UMC UMC UMC = 554D4320 554D4320 554D4320}

function is_umc : Boolean; assembler;
asm
  db 66h;xor   ax,ax
  db 0fh,0A2h {cpuid}
  db 66h;cmp   ax,1;dw 0000h
         jl @LNoUMC
  db 66h;sub   bx,4D55h;dw 2043h
  db 66h;sub   dx,4D55h;dw 2043h
  db 66h;sub   cx,4D55h;dw 2043h
  db 66h;or    bx,cx
  db 66h;or    bx,dx
  db 0Fh,94h,0C0h{setz al}
         jmp @LUMCExit
@LNoUMC:
         mov  ax,0
@LUMCExit:
end;

{ NexGenDriven = 4E657847 656E4472 6976656E}

function is_nexgen : Boolean; assembler;
asm
  db 66h;xor   ax,ax
  db 0fh,0A2h {cpuid}
  db 66h;cmp   ax,1;dw 0000h
         jl @LNoNexGen
  db 66h;sub   bx,654Eh;dw 4778h
  db 66h;sub   dx,6E65h;dw 7244h
  db 66h;sub   cx,7669h;dw 6E65h
  db 66h;or    bx,cx
  db 66h;or    bx,dx
  db 0Fh,94h,0C0h{setz al}
        jmp @LNexGenExit
@LNoNexGen:
         mov   ax,0
@LNexGenExit:
end;

{ CentaurHauls = 43656E74 61757248 61756C73
  We actually don't use the CPUid string here, we check if CPUid function
  C000:0000 is available and supported.
}

function is_centaur : Boolean; assembler;
asm
  db 66h;mov   ax,0000h;dw 0C000h
  db 0fh,0A2h {cpuid}
  db 66h;sub   ax,0000h;dw 0C000h
  db 0Fh,94h,0C0h{setz al}
end;

{ Capabilities support check
  The different capabilites of processors are mostly detected by examining
  CPUid function 1, although some require additional work.
}

function fpu_supported : Boolean; assembler;
asm
  fninit
  mov   ax,05A5Ah
  fnstsw ax
  or    ax,ax
  db 0Fh,94h,0C0h{sete al}
end;

{ check if the cpuid instruction is available }
function cpuid_supported : Boolean; assembler;
asm
  db 66h;pushf
  db 66h;pushf                  { Get original EFLAGS}
  db 66h;pop   ax
  db 66h;mov   cx,ax
  db 66h;xor   ax,0;dw 20h      { Flip id bit in eflags }
  db 66h;push  ax               { Save new eflags value on stack }
  db 66h;popf                   { Replace current eflags value }
  db 66h;pushf                  { Get new eflags }
  db 66h;pop   ax               { Store new eflags in eax }
  db 66h;popf                   { Restore original eflags }
  db 66h;xor   ax,cx            { Can not toggle ID bit}
  db 0Fh,95h,0C0h {setnz al}
end;

{ check for rdtsc support by examining bit 4 returned in %edx by cpuid
  function 1
}
function rdtsc_supported : Boolean; assembler;
asm
  db 66h;mov   ax,1;dw 0000h
  db 0Fh,0A2h {cpuid}
         test  dx,10h
  db 0Fh,95h,0C0h {setnz al}
end;

{ check for MMX support by examining bit 23 returned in %edx by the cpuid
  function 1
}
function mmx_supported : Boolean; assembler;
asm
  db 66h;xor   ax,ax
  db 0Fh,0A2h {cpuid}
  db 66h;cmp   ax,1;dw 0000h
         jl    @LMMXFail
  db 66h;mov   ax,1;dw 0000h
  db 0Fh,0A2h {cpuid}
  db 66h;and   dx,0000h; dw 0080h
  db 0Fh,95h,0C0h {setnz al}
  jmp @LMMXExit
@LMMXFail:
   xor   al,al
@LMMXExit:
end;

{ check for eMMX support. It's obtained by checking bit 24 returned in edx
  of cpuid function 8000:0001 (functions >= 8000:0000 are only supported by
  AMD and Cyrix processors
}
function emmx_supported : Boolean; assembler;
asm
  db 66h;mov   ax,0000h;dw 8000h
  db 0Fh,0A2h  {cpuid}
  db 66h;cmp   ax,0001h;dw 8000h
         jb    @LeMMXFail
  db 66h;mov   ax,0001h;dw 8000h
  db 0Fh,0A2h  {cpuid}
  db 66h;test  dx,0000h;dw 0100h
  db 0Fh,95h,  0C0h {setnz al}
         jmp   @LeMMXExit
@LeMMXFail:
         xor   ax,ax
@LeMMXExit:
end;

{ check for 3dnow! support, by examing bit 31 returned in edx of cpuid
  function 8000:0001 (only supported by AMD and some Cyrix processors)
}
function _3dnow_supported : Boolean; assembler;
asm
  db 66h;mov   ax,0000h;dw 8000h
  db 0Fh,0A2h  {cpuid}
  db 66h;cmp   ax,0001h;dw 8000h
          jb    @L3DnowFail
  db 66h;mov   ax,0001h;dw 8000h
  db 0Fh,0A2h  {cpuid}
  db 66h;test  dx,0000h;dw 8000h
  db 0Fh,95h,  0C0h {setnz al}
         jmp   @L3DnowExit
@L3DnowFail:
          xor   ax,ax
@L3DnowExit:
end;

{ check for Intel's Katmai New Instructions (ISSE) set. This is bit 25 in edx
  returned by cpuid function 1.
}
function kni_supported : Boolean; assembler;
asm
  db 66h;mov   ax,0001h;dw 0000h
  db 0Fh,0A2h  {cpuid}
  db 66h;test  dx,0000h;dw 0200h
  db 0Fh,95h,  0C0h {setnz al}
end;

{ check if the 96 bit serial number is present and available. PIII+ only
}

function sernum_supported : Boolean; assembler;
asm
  db 66h;mov   ax,0001h;dw 0000h
  db 0Fh,0A2h  {cpuid}
  db 66h;test  dx,0000h;dw 0004h
  db 0Fh,95h,  0C0h {setnz al}
end;

{ check out stepping, model, family via CPUid instruction }

function get_stepping : Byte; assembler;
asm
  db 66h;mov   ax,0001h;dw 0000h
  db 0Fh,0A2h  {cpuid}
         and   al,0Fh
end;

function get_model : Byte; assembler;
asm
  db 66h;mov   ax,0001h;dw 0000h
  db 0Fh,0A2h  {cpuid}
         shr   ax,4
         and   al,0Fh
end;

function get_family : Byte; assembler;
asm
  db 66h;mov   ax,0001h;dw 0000h
  db 0Fh,0A2h  {cpuid}
         shr   ax,8
         and   al,0Fh
end;

{ copies the vendor string returned by cpuid function 0 returned in ebx, edx,
  and ecx
}

procedure get_cpu_string(var c : string); assembler;
asm
  db 66h;xor   ax,ax
  db 0Fh,0A2h  {cpuid}
         les   di,c
  mov  es:[di].byte,12
  db 66h;mov  es:[di+1],bx
  db 66h;mov  es:[di+5],dx
  db 66h;mov  es:[di+9],cx
end;

{ returns the processors unique 96 bit serial number which is returned by
  cpuid function 1 in eax (top 32 bit) and cpuid function 3 in ecx (bottom
  32 bit) and edx (middle 32 bit) respectively
}

procedure get_sernum(var s : serial_number); assembler;
asm
  les   di,s
  db 66h;mov   ax,0001h;dw 0000h
  db 0Fh,0A2h  {cpuid}
  db 66h;mov   es:[di],ax
  db 66h;mov   ax,0003h;dw 0000h
  db 0Fh,0A2h  {cpuid}
  db 66h;mov   es:[di+4],dx
  db 66h;mov   es:[di+8],cx
end;


{ CPU detection routines
  These routines are executed when no cpuid instruction was found.
}

function is_386 : Boolean; assembler;
asm
{ distuingish between 386/486 by examining the ability to preserve the AC bit
  of the flags register. (386 = no, 486 = yes)  }
         mov   bx,sp           { save current stack pointer }
         and   sp,0FFFCh       { align stack to avoid AC fault }
  db 66h;pushf
  db 66h;pop   ax
  db 66h;mov   cx,ax
  db 66h;xor   ax,0000h;dw 4h  { xor AC bit in EFLAGS }
  db 66h;push  ax
  db 66h;popf
  db 66h;pushf
  db 66h;pop   ax
  db 66h;xor   ax,cx           { Is AC toggled ? }
  db 0Fh,94h,0C0h{setz al}
         and   sp,0FFFCh       { align stack to prevent AC fault }
  db 66h;push  cx
  db 66h;popf                  { restore original AC bit }
         mov   sp,bx           { restore original stack pointer }
end;

function is_486 : Boolean; assembler;
asm
{ Distinguish between the i486 and Pentium by the ability to set the ID flag
  in the EFLAGS register. If the ID flag is set, then we can use the CPUID
  instruction to determine the final version of the chip. Otherwise we
  simply have an 80486.}
  db 66h;pushf                { Get original EFLAGS}
  db 66h;pop   ax
  db 66h;mov   cx,ax          { save EFLAGS }
  db 66h;xor   ax,0000;dw 20h { Flip ID bit in EFLAGS}
  db 66h;push  ax             { Save new EFLAGS value on stack}
  db 66h;popf                 { Replace current EFLAGS value}
  db 66h;pushf                { Get new EFLAGS}
  db 66h;pop   ax             { Store new EFLAGS in EAX}
  db 66h;xor   ax,cx          { Toggle ID bit}
  db 0Fh,94h,0C0h{setz al}
  db 66h;push  cx             { restore old EFLAGS }
  db 66h;popf
end;


{ helper functions to decide between various processor types }

function is_Pentium : Boolean;
begin
     if (cpuid_supported and (get_family = 5)) then is_Pentium := true
     else is_Pentium := false;
end;

function is_PentiumPro : Boolean;
begin
     if (cpuid_supported and (get_family = 6) and (get_stepping < 2)) then is_PentiumPro := true
     else is_PentiumPro := false;
end;

function is_PentiumII : Boolean;
begin
     if (cpuid_supported and (get_family = 6) and (get_stepping > 1)) then is_PentiumII := true
     else is_PentiumII := false;
end;

{ main cpu detection routine which uses all the procedures above  }

procedure get_cpu_information(var c : cpu_rec);
begin
     with (c) do begin
          { set default values }

          model := $ff;
          stepping := $ff;
          cpu_id := 'Unknown';
          family := cFAMILY_UNKNOWN;
          manufacturer := cMFC_OTHER;
          capabilities := 0;
          sernum[0] := 0;
          sernum[1] := 0;
          sernum[2] := 0;

          { check capabilities first }

          { fpu availability check }
          if (fpu_supported) then capabilities := capabilities or cHAVE_FPU;
          { cpuid instruction check }
          if (cpuid_supported) then begin
             capabilities := capabilities or cHAVE_CPUid;

             { get other information obtained by the cpuid instruction }
             get_cpu_string(cpu_id);
             model := get_model;
             stepping := get_stepping;
             family := get_family;

             { RDTSC needs CPUid }
             if (rdtsc_supported) then capabilities := capabilities or cHAVE_RDTSC;
             { MMX support check }
             if (mmx_supported) then capabilities := capabilities or cHAVE_MMX;

             { (e)MMX, 3Dnow! and KNI only available when cpuid is supported }

             { only Cyrix' Media GXm processors may have extended MMX }
             if ((is_cyrix and is_PentiumII) and (emmx_supported)) then
                capabilities := capabilities and cHAVE_eMMX;
             { only AMD processors with model > 0 or Cyrix GXm (Media GX) processors
               may support 3Dnow! only }
             if (((is_amd and (model > 0)) or (is_cyrix and is_PentiumII)) and (_3dnow_supported)) then
                capabilities := capabilities or cHAVE_3DNOW;
             { KNI (ISSE) check }
             if (kni_supported) then capabilities := capabilities or cHAVE_KNI;
             { Serial number check }
             if (sernum_supported) then begin
                capabilities := capabilities or cHAVE_SERNUM;
                get_sernum(sernum);
             end;
          end else begin
              { get family without cpuid }
              if (is_386) then family := cFAMILY_I386 else
              if (is_486) then family := cFAMILY_I486 else
              family := cFAMILY_PENTIUM;
          end;

          { measure clock rate of CPU }
          clock_rate := get_cpu_clockrate(family);
          { and find suitable rating }
          rating := get_cpu_rating(clock_rate);

          { manufacturer check only done when cpuid is supported }
          if (cpu_supported(cHAVE_CPUID)) then begin
             if (is_centaur) then manufacturer := cMFC_CENTAUR else
             if (is_amd) then manufacturer := cMFC_AMD else
             if (is_cyrix) then manufacturer := cMFC_CYRIX else
             if (is_intel) then manufacturer := cMFC_INTEL else
             if (is_nexgen) then manufacturer := cMFC_NEXGEN else
             if (is_umc) then manufacturer := cMFC_UMC else
             manufacturer := cMFC_OTHER;
          end;
     end;
end;

{ check capabilities field of cpu - record and return true if all conditions
  are met by the current processor
}

function cpu_supported(capability : DWord) : Boolean;
begin
     cpu_supported := ((this_cpu.capabilities and capability) = capability);
end;

{ clockrate detection
  There are two methods how speed detection is done: Either by using the
  rdtsc instruction if available or the PIT timer.

  The rdtsc speed detection is done as follows : The number of instructions
  executed in a constant time is easily deducted by using the values the
  rdtsc returns. This is done until three subsequent measure runs are within
  a tolerance of +- 1 Mhz or a specific number of runs were done. The
  frequency then calculates by dividing the highly accurate number of ticks
  elapsed in the timing interval by the length of the timing interval.

  On the other hand, when no rdtsc is available (on all 386, most 486, some
  older Pentiums and Cyrix' 6x86) the time of a loop where the tick count is
  known is run some thousand times and the time span to complete this is
  measured. This is done about 10 times and the run which used least time is
  taken for frequency calculation. The frequency of the CPU then calculates by
  dividing the calculated number of ticks the loop runs take by the time
  measured. Due to the nature of this method it is not avoidable that the
  frequency may be a bit inaccurate, especially because some newer steppings
  of 386, 486 or Pentiums are a bit faster at executing the timing loop than
  expected. Another source for errors are those turbo switches, because when
  disabled, the CPUs actually seem to execute some instructions faster than
  displayed in the opcode lists.
  (e.g. A 486DX2-66 with turbo switch enabled is measured at a speed of 67 Mhz
  which is right, but when the turbo switch is disabled, timing results
  _always_ give 38 Mhz... another example is a 486DX-33: With turbo the unit
  times for 33 Mhz, but when it is disabled, it measures 31 Mhz instead of
  30...)

  Additionally one of the above methods is executed about five times and then
  the result is an average of these runs.
}


{ return the contents of the time stamp counter }

procedure get_rdtsc_value(var val : comp); assembler;
asm
  db 0Fh,31h{rdtsc}
         les   di,val
  db 66h;mov   es:[di],ax
  db 66h;mov   es:[di+4],dx
end;

{ return the current PIT value }

procedure get_pit_value(var val : comp); assembler;
asm
  db 66h;xor   ax,ax
         cli
         out   43h,al
         mov   es,[seg0040]
  db 66h;mov   dx,es:[6Ch]
         in    al,40h
  db 0EBh,00h,0EBh,00h,0EBh,00h
         mov   ah,al
         in    al,$40
  db 0EBh,00h,0EBh,00h,0EBh,00h
         xchg  ah,al
         neg   ax
  db 66h;xor   di,di
         mov   di,ax
         sti
  db 66h;mov   bx,0000h;dw 1
  db 66h;mov   ax,dx
  db 66h;xor   dx,dx
  db 66h;mul   bx
  db 66h;add   ax,di
  db 66h;adc   dx,0
         les   di,val
  db 66h;mov   es:[di],ax
  db 66h;mov   es:[di+4],dx
end;

{ measure CPU speed by means of the rdtsc instruction }

function get_rdtscspeed : Longint;

const MIN_TRIES = 3;  { number of minumum tries }
      MAX_TRIES = 20; { number of maximume tries issued }

      TOLERANCE = 1; { the tolerance in Mhz three subsequent
                       measurement results may differ }

var freq : array[0..2] of comp; { storage for the 3 subsequent measurements }
    time0, time1,        { start and endtime (PIT values) }
    stamp0, stamp1,      { start and endcycles (rdtsc values) }
    cycles, ticks : comp;{ used for frequency calculation }
    total : comp;        { total amount of frequency calculated from the 3
                           subsequent runs }
    dummy,
    tries : DWord;       { current number of tries }

begin
     { initialize variables }
     for dummy := 0 to 2 do freq[dummy] := 0;
     tries := 0;
     time0 := 0; time1 := 0; stamp0 := 0; stamp1 := 0;
     cycles := 0; ticks := 0; total := 0;
     repeat
           { here to avoid overflows (e.g. for FPC) }
           for dummy := 0 to 2000 do ;
           { increment number of tries }
           tries := tries + 1;

           { rotate frequency values in storage }
           freq[2] := freq[1];
           freq[1] := freq[0];
           { this accounts for overhead later }
           get_pit_value(time0);
           time1 := time0;
           while ((time1-time0) < 50) do begin
                 get_pit_value(time1);
                 get_rdtsc_value(stamp0);
           end;

           { again, prevent overflows (e.g. for FPC) }
           for dummy := 0 to 2000 do ;

           { here's the actual measurement }
           time0 := time1;
           get_pit_value(time0);
           get_rdtsc_value(stamp0);
           { time for about 5 ms }
           while ((time1-time0) < 5000) do begin
                 get_pit_value(time1);
                 get_rdtsc_value(stamp1);
           end;

           cycles := stamp1 - stamp0;
           ticks := ((time1 - time0) * 1000000) / PIT_RATE;
           { calculate frequency }
           freq[0] := cycles / ticks;
           total := freq[0] + freq[1] + freq[2];

           { if we've made enough tries, exit }
           if (tries > MAX_TRIES) then break;
           { loop until three subsequent measurements are within tolerance }
     until ((abs((3*freq[0])-total) <= TOLERANCE) and
            (abs((3*freq[1])-total) <= TOLERANCE) and
            (abs((3*freq[2])-total) <= TOLERANCE));
     { return speed value }
     get_rdtscspeed := round(total / 3);
end;

{ measure CPU speed by means of timing the execution of known loop }

const NUMLOOPS = 20000;  { number of loops executed }

function get_bsfspeed(cycles : DWord) : DWord;

{ the timed loop }

procedure runBSFLoop(iter : DWord); assembler;
asm
  db 66h;mov   bx,ITER.word
  db 66h;mov   ax,0000h;dw 8000h
@LBSFLoop:
  db 66h,0Fh,0BCh,0C8h {bsf ecx,eax}
  db 66h;dec   bx
  jnz @LBSFLoop
end;

const SAMPLINGS = 10;   { number of times the loop is timed at one run }

var t0, t1,     { start and endtime }
    lowest,     { lowest time measured }
    ticks,
    current : Comp;
    i : DWord;

begin
     { initialize lowest to a huge value }
     lowest := $7fffffff;
     { time loop }
     for i := 0 to SAMPLINGS-1 do begin
         get_pit_value(t0);
         runBSFLoop(NUMLOOPS);
         get_pit_value(t1);
         current := abs(t1 - t0);
         if (current < lowest) then lowest := current;
     end;
     { scale the PIT frequency to microsecond values }
     ticks := ((lowest * 1000000) / PIT_RATE);
     get_bsfspeed := round(cycles / ticks);
end;


{ the clock-rate routine which chooses between the above two methods and
  averages about some measurements
}

function get_cpu_clockrate(family : Byte) : DWord;

const RUNS = 5;       { the number of times timing is done }

      { the number of cycles needed for the different CPU types for the
        non-rdtsc speed detection }
      bsf_cycles : array[cFAMILY_I386..(cFAMILY_UNKNOWN-1)] of DWord =
                 (115, 47, 43, 38);

var times : Longint;
    avgspeed : DWord;
    speed  : DWord;
    use_rdtsc : Boolean;

begin
     avgspeed := 0;
     use_rdtsc := cpu_supported(cHAVE_CPUID or cHAVE_RDTSC);
     for times := 1 to RUNS do begin
         if (use_rdtsc) then speed := get_rdtscspeed
         else speed := get_bsfspeed(NUMLOOPS*bsf_cycles[family]);
         avgspeed := avgspeed + speed;
     end;
     get_cpu_clockrate := (avgspeed div RUNS);
end;


{ cpu rating detection }


function get_cpu_rating(clock : DWord) : DWord;
var i : Longint;
    prev_abs : DWord;
begin
     prev_abs := abs(known_ratings[0]-clock); i := 1;
     while (prev_abs > abs(known_ratings[i] - clock)) do begin
           prev_abs := abs(known_ratings[i] - clock);
           i := i + 1;
     end;
     get_cpu_rating := (known_ratings[(i-1)]);
end;

{ family string output }

function get_family_string(f : DWord) : string;

      { the family string identifiers returned for the different cpu types }
const family_strings : array[cFAMILY_I386..cFAMILY_UNKNOWN] of string[12] =
                     ('80386',
                      '80486',
                      'Pentium',
                      'PentiumII',
                      '>PentiumII');
begin
     if (f in [cFAMILY_I386..(cFAMILY_UNKNOWN-1)]) then
        get_family_string := family_strings[f]
     else
         get_family_string := family_strings[cFAMILY_UNKNOWN];
end;

function get_mfc_string(m : DWord) : string;

const mfc_strings : array[cMFC_AMD..cMFC_OTHER] of string[12] =
                    ('AMD',
                     'Centaur',
                     'Cyrix',
                     'Intel',
                     'NexGen',
                     'UMC',
                     'Other');

begin
     if (m in [cMFC_AMD..(cMFC_OTHER-1)]) then
        get_mfc_string := mfc_strings[m]
     else
         get_mfc_string := mfc_strings[cMFC_OTHER];
end;

function hexstr(l : longint;width : word) : string;
const
  hexs : array[0..15] of char=
  ('0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F');
var
  s : string[10];
  i : longint;
begin
  s := '';
  for i := 0 to width-1 do
  begin
    s := hexs[l and 15]+s;
    l := l shr 4;
  end;
  hexstr := s;
end;

function get_sernum_string(ser : serial_number) : string;
var i : longint;
    s : string;
begin
     s := '';
     for i := 0 to 2 do
         s := s + hexstr(hi(ser[i]), 4) + '-' + hexstr(lo(ser[i]), 4) + '-';
     delete(s, length(s), 1);
     get_sernum_string := s;
end;


{****************************************************************************}
{                                                                            }
{ Timer Object                                                               }
{                                                                            }
{ - works on any platform (uses dos unit)                                    }
{ - timing resolution of 100 ms                                              }
{                                                                            }
{****************************************************************************}

constructor tTimer.Init;
begin
     Reset;
end;

destructor tTimer.Done;
begin
end;

procedure tTimer.Reset;
begin
     tmStart := 0;
     tmElapsed := 0;
     timing := false;
end;

procedure tTimer.Start;
begin
     timing := true;
     tmStart := ReadTime;
end;

procedure tTimer.Stop;
begin
     tmElapsed := tmElapsed + (ReadTime - tmStart);
     timing := false;
     tmStart := 0;
end;

function tTimer.Count : Comp;
begin
     Count := tmElapsed;
end;

function tTimer.Lap : Comp;
begin
     lap := tmElapsed + (readtime - tmStart)
end;

function tTimer.Resolution : Real;
begin
     resolution := 1e-2;
end;

function tTimer.IsTiming : Boolean;
begin
     IsTiming := timing;
end;

procedure tTimer.Delay(time : DWord);
begin
     time := round(time/resolution/1000);
     Start;
     while (lap < time) do ;
     Stop;
     Reset;
end;

procedure tTimer.WaitFor(time : DWord);
begin
     time := round(time/resolution/1000);
     while (lap < time) do ;
     Stop;
     Reset;
end;

function tTimer.ReadTime : Comp;
var hour, minute, second, sec100 : Word;
begin
     gettime(hour, minute, second, sec100);
     ReadTime := (sec100 + (second + (minute + hour * 60) * 60) * 100);
end;

{****************************************************************************}
{                                                                            }
{ ZenTimer Timer Object                                                      }
{                                                                            }
{ - GO32V2 only                                                              }
{ - timing resolution of 1 microsecond                                       }
{                                                                            }
{****************************************************************************}

constructor tZTimer.Init;
begin
     use_rdtsc := cpu_supported(cHAVE_CPUID or cHAVE_RDTSC);
     if (use_rdtsc) then rate := this_cpu.clock_rate
     else rate := PIT_RATE;
     Reset;
end;

destructor tZTimer.Done;
begin
end;

function tZTimer.Count : Comp;
begin
     Count := TicksTo_usec(tmElapsed);
end;

function tZTimer.Lap : Comp;
begin
     lap := TicksTo_usec(readtime - tmStart)
end;

function tZTimer.Resolution : Real;
begin
     resolution := 1e-6;
end;

function tZTimer.TicksTo_usec(t : Comp) : Comp;
begin
     if (use_rdtsc) then
        TicksTo_usec := t / rate
     else
         TicksTo_usec := t * 1000000 / rate;
end;

function tZTimer.ReadTime : Comp;
var q : Comp;
begin
     if (use_rdtsc) then
        get_RDTSC_value(q)
     else
         get_PIT_value(q);
     ReadTime := q;
end;

{****************************************************************************}
{                                                                            }
{ Ultra Long Timer Object                                                    }
{                                                                            }
{ - GO32V2 only                                                              }
{ - timing resolution of 1/18.2 second                                       }
{                                                                            }
{****************************************************************************}


constructor ULZTimer.Init;
begin
end;

destructor ULZTimer.Done;
begin
end;

function ULZTimer.Resolution : Real;
begin
     Resolution := $10000/PIT_RATE;
end;
            
function ULZTimer.ReadTime : Comp;
var t : DWord;
begin
     asm
              cli
              mov   es,[seg0040]
       db 66h;mov   ax,es:[6Ch]
              sti
       db 66h;mov t.word,ax
     end;
     ReadTime := t;
end;

begin
     { brute force PIT reset }
     asm
       mov ax,34h
       out 43h,al
       xor al,al
       out 40h,al
       out 40h,al
     end;
     { get cpu info for standard cpu record }
     get_cpu_information(this_cpu);
     {$IFDEF DEBUG}
     if (cpu_supported(cHAVE_RDTSC)) then Writeln(stderr, 'CPU : Using RDTSC for microsecond timing')
     else Writeln(stderr, 'CPU : Using PIT for microsecond timing');
     Writeln(stderr);
     {$ENDIF}
end.

begin
end.